/* Map a shared object's segments into hugepage.  Generic version.
   Copyright (C) 1995-2021 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <http://www.gnu.org/licenses/>.  */

#include <dl-load.h>

#define SHFIT_2MB 21
#define SIZE_2MB 0x200000
#define MASK_2MB 0x1FFFFF
#define THRESHOLD 16

/*
 * Find the first PT_LOAD segment with execute permission
 */
static __always_inline const struct loadcmd *
_find_exec_segment(const struct loadcmd loadcmds[], size_t nloadcmds)
{
  const struct loadcmd *c = loadcmds;

  while (c < &loadcmds[nloadcmds])
    {
      if (c->prot & PROT_EXEC)
        return c;
      c++;
    }
  return NULL;
}

static __always_inline void *
__mmap_reserved_area(const struct loadcmd loadcmds[], size_t nloadcmds,
                    size_t *maparealen)
{
  const struct loadcmd * c = loadcmds;
  *maparealen = 0;

  /*
   * Consider maybe-existing hole between PT_LOAD segments
   */
  while (c < &loadcmds[nloadcmds])
    {
      // c is not the last loadcmd
      if (c + 1 < &loadcmds[nloadcmds]) {
        *maparealen += ALIGN_UP((c + 1)->mapstart - c->mapstart, SIZE_2MB);
      } else {
        *maparealen += ALIGN_UP((c->mapend > c->allocend ? c->mapend : c->allocend), SIZE_2MB) -
                      ALIGN_DOWN(c->mapstart, SIZE_2MB);
      }
      c++;
    }

  /*
   * Get 2MB aligned contiguous va space
   * This va space can not be munmap in case of multi thread dlopen concurrently
   */
  void *map_area_start = __mmap(0, *maparealen, PROT_NONE,
    MAP_PRIVATE|MAP_ANONYMOUS|MAP_HUGETLB|MAP_NORESERVE|(SHFIT_2MB << MAP_HUGE_SHIFT), -1, 0);
  if (__glibc_unlikely (map_area_start == MAP_FAILED))
    return MAP_FAILED;

  /*
   * Remap 2MB aligned contiguous va space into 4KB contiguous va space
   * to avoid the tedious work of splitting hugepage into 4KB page
   */
  if (__glibc_unlikely(__mmap(map_area_start, *maparealen, PROT_NONE,
                              MAP_PRIVATE|MAP_ANONYMOUS|MAP_FIXED, -1, 0)
                        == MAP_FAILED))
    {
      goto unmap_reserved_area;
    }
  return map_area_start;

unmap_reserved_area:
  __munmap(map_area_start, *maparealen);
  return MAP_FAILED;
}

/*
 * With traditionad -z nosperate-code, GNU ld defaults to a RX/R/RW program header layout,
 * With -z sperate-code(defallt on Linux/x86 form binutils 2.31 onwards), GNU ld defaults to a R/RX/R/RW program header layout
 *
 * ld.lld defaults to R/RX/RW(RELRO)/RW(non-RELRO), with --rosegment, ld.lld uses RX/RW(RELRO)/RW(non-RELRO)
 * LLVM LLD put RELRO in the sperate PT_LOAD(PT_GNU_RELRO(.data.rel.ro .bss.rel.ro)); PT_LOAD(.data. .bss)
 * LLVM LLD default RELRO filesz < RELRO memsz, and LLVM can not keep compatible with GNU LD: RELRO fiesz  == RELRO memsz
 *
 * hugepage feature of dynamic library is consistent with the default glibc dynamic library loading process: mmap file part first, then map anonymous part
 * As GNU LD RELRO fiesz  == RELRO memsz, treat RELRO as file part,
 * but RELRO filesz < RELRO memsz which will cause wrong relro len caculation
 * so just mmap LLVM LLD RELRO PT_LOAD as normal page to avoid GNU LD relro len caculation
 */
static __always_inline void
_llvm_ld_relro_hp_fallback(const ElfW(Phdr) *phdr, const ElfW(Phdr) *relro_ph,
                     struct link_map *l, char *hp_bitmap)
{
  const ElfW(Phdr) *ph;
  char *hp_loadcmd;
  int i = 0;

  // all the PT_LOADs do not contain RELRO, do nothing
  if (relro_ph == NULL)
    return;

  for (ph = phdr, hp_loadcmd = hp_bitmap; ph < &phdr[l->l_phnum]; ++ph)
    if (ph->p_type == PT_LOAD)
      {
        // relro_ph->p_vaddr + relro_ph->p_memsz > ph->p_vaddr + ph->p_memsz
        //  --> llvm.ld no padding bits, but will report error in _dl_protect_relro
        if (ph->p_vaddr <= relro_ph->p_vaddr && relro_ph->p_vaddr < ph->p_vaddr + ph->p_memsz &&
           relro_ph->p_filesz < relro_ph->p_memsz)
          {
            if (__glibc_unlikely (GLRO(dl_debug_mask) & DL_DEBUG_FILES))
              _dl_debug_printf("PT_LOAD[%d] contains LLVM LD layout style RELRO, this segment use normal page\n", i);

            *hp_loadcmd = 0;
            break;
          }
         else
          {
            hp_loadcmd++;
            i++;
          }
      }

  return;
}

/*
 * | other sections | RELRO | non-RELRO |
 * If RELRO is not at the beginning of the RW segment, the beginning position
 * will also be mapped using 4KB page, so count the begining other sections in relro len
 * RELRO is default at the begining position of GNU LD and ld.lld
 * relro_len is GLRO(dl_pagesize) aligned
 */
static __always_inline size_t
_get_relro_len(struct link_map *l, const struct loadcmd *c)
{
  size_t relro_len = 0;

  if (l->l_relro_size)
    {
      ElfW(Addr) relro_start = ALIGN_DOWN(l->l_relro_addr, GLRO(dl_pagesize));
      ElfW(Addr) relro_end = ALIGN_DOWN((l->l_relro_addr + l->l_relro_size), GLRO(dl_pagesize));
      if (c->mapstart <= relro_start && relro_start < c->allocend)
        relro_len = relro_end - c->mapstart;
    }
  return relro_len;
}

/*
 * the alignment stands for the size of page which is to be cleared to zero
 */
static __always_inline const char *
_zero_tail_page(const struct loadcmd *c, ElfW(Addr) zero, ElfW(Addr) zeropage,
                size_t alignment)
{
  if (__glibc_unlikely ((c->prot & PROT_WRITE) == 0))
    {
      /* Dag nab it.  */
      if (__mprotect ((caddr_t) ALIGN_DOWN(zero, alignment), alignment,
                      c->prot|PROT_WRITE) < 0)
        return DL_MAP_SEGMENTS_ERROR_MPROTECT;
    }
    memset ((void *) zero, '\0', zeropage - zero);
    if (__glibc_unlikely ((c->prot & PROT_WRITE) == 0))
      __mprotect ((caddr_t) ALIGN_DOWN(zero, alignment), alignment, c->prot);
    return NULL;
}

static __always_inline const char *
_mmap_remain_zero_page(ElfW(Addr) zeropage, ElfW(Addr) zeroend, int prot, char hp_flag)
{
  ElfW(Addr) hp_start = ALIGN_UP(zeropage, SIZE_2MB);
  size_t len = 0, mod = 0;
  caddr_t mapat;

  if (zeroend > hp_start && zeroend - hp_start >= SIZE_2MB)
    {
      len = zeroend - hp_start;
      mod = len % SIZE_2MB;
    }
  else
    hp_start = 0;

  if (hp_start == 0)
    {
      mapat = __mmap((caddr_t) zeropage, zeroend - zeropage, prot,
                        MAP_ANON|MAP_PRIVATE|MAP_FIXED, -1, 0);
      if (__glibc_unlikely (mapat == MAP_FAILED))
        return DL_MAP_SEGMENTS_ERROR_MAP_ZERO_FILL;
      return NULL;
    }

  if (hp_start - zeropage > 0)
    {
      mapat =  __mmap((caddr_t) zeropage, hp_start - zeropage,
                              prot, MAP_ANON|MAP_PRIVATE|MAP_FIXED, -1, 0);
      if (__glibc_unlikely (mapat == MAP_FAILED))
        return DL_MAP_SEGMENTS_ERROR_MAP_ZERO_FILL;
    }

  if (__glibc_unlikely (GLRO(dl_debug_mask) & DL_DEBUG_FILES))
    _dl_debug_printf("\t\t\t=> %s mmap anonymous hugepage: [%lx-%lx)\n", __func__, hp_start, hp_start + len - mod);

  if (hp_flag == DL_HUGEPAGE_LARGE_IN)
  {
    mapat = __mmap((caddr_t) hp_start, len - mod, prot,
		    MAP_FIXED|MAP_PRIVATE|MAP_ANONYMOUS|MAP_HUGETLB|(SHFIT_2MB << MAP_HUGE_SHIFT),
                    -1, 0);

    if (__glibc_unlikely (mapat == MAP_FAILED))
      return DL_MAP_SEGMENTS_ERROR_MAP_ZERO_FILL;
  } else {
    mapat = __mmap((caddr_t) hp_start, len - mod, prot,
                    MAP_FIXED|MAP_PRIVATE|MAP_ANONYMOUS, -1, 0);

    if (__glibc_unlikely (mapat == MAP_FAILED))
      return DL_MAP_SEGMENTS_ERROR_MAP_ZERO_FILL;

    (void) madvise((void *)hp_start, len - mod, MADV_HUGEPAGE);
  }

  if (mod > 0)
    {
      if (__glibc_unlikely (GLRO(dl_debug_mask) & DL_DEBUG_FILES))
        _dl_debug_printf("\t\t\t=> %s mmap anonymous tail normal page: [%lx-%lx)\n", __func__,
                        hp_start + len - mod, hp_start + len - mod + ALIGN_UP(mod, GLRO(dl_pagesize)));
      mapat =__mmap((caddr_t)(hp_start + len - mod), mod, prot,
                      MAP_ANON|MAP_PRIVATE|MAP_FIXED, -1, 0);
    if (__glibc_unlikely (mapat == MAP_FAILED))
      return DL_MAP_SEGMENTS_ERROR_MAP_ZERO_FILL;
    }

  return NULL;
}

/*
 * memsz_len records the remain memsiz part
 */
static __always_inline const char *
_mmap_segment_memsz(struct link_map *l, const struct loadcmd * c,
                ElfW(Addr) mapstart, size_t extra_len, size_t *memsz_len, char hp_flag)
{
  const char * errstring = NULL;

  /* Extra zero pages should appear at the end of this segment,
     after the data mapped from the file.   */
  ElfW(Addr) zero, zeroend, zeropage;

  zero = mapstart + c->dataend - c->mapstart;
  zeroend = mapstart + c->allocend - c->mapstart;
  zeropage = ALIGN_UP(zero, GLRO(dl_pagesize));
  size_t alignment = GLRO(dl_pagesize);
  *memsz_len = 0;

  /*
   * no matter what the extra space consists of:
   * 1. all the extra space is initialized data area (MemSiz > FileSiz)
   * 2. initialized data area and hole
   * 3. all the extra space is hole (MemSiz == FileSiz)
   *
   * the extra space just needs to be set zero, for the initialized data area, it's
   * initialized to zero; for the hole area, it's initialized to invalid instruction
   */
  if (extra_len > 0)
    {
      if (__glibc_unlikely(zeropage == ALIGN_UP(zero, SIZE_2MB) ||
                            zeropage + extra_len != ALIGN_UP(zero, SIZE_2MB)))
        return DL_MAP_SEGMENT_ERROR_EXTRA_SIZE;

      zeropage = ALIGN_UP(zero, SIZE_2MB);
      alignment = SIZE_2MB;
    }
  else
    {
      /*
       * extra_len = 0, _mmap_segment_filesz just mmap segment's FileSiz part,
       * here, it needs to zero tail page [FileSiz end, tail page end) part
       */
      if (c->allocend <= c->dataend)
        return NULL;

      if (ALIGN_UP(zero, GLRO(dl_pagesize)) == ALIGN_UP(zero, SIZE_2MB) &&
            (zeropage - (mapstart + _get_relro_len(l, c)) >= SIZE_2MB))
        {
          alignment = SIZE_2MB;
        }

      if (zeroend < zeropage)
        zeropage = zeroend;
    }

  if (zeropage > zero)
    {
        if (__glibc_unlikely (GLRO(dl_debug_mask) & DL_DEBUG_FILES))
          _dl_debug_printf("\t\tzero tail page [%lx-%lx) which contains hole area length: 0x%lx\n",
                          zero, zeropage, zeropage > ALIGN_UP(zero, GLRO(dl_pagesize)) ?
                          zeropage - ALIGN_UP(zero, GLRO(dl_pagesize)) : 0);
        errstring = _zero_tail_page(c, zero, zeropage, alignment);
        if (errstring != NULL)
          return errstring;
    }

  if (zeroend > zeropage)
    {
      *memsz_len = ALIGN_UP(zeroend, GLRO(dl_pagesize)) - zeropage;
      if (__glibc_unlikely (GLRO(dl_debug_mask) & DL_DEBUG_FILES))
        _dl_debug_printf("\t\tmmap remain anonymous page [%lx-%lx)\n", zeropage, ALIGN_UP(zeroend, GLRO(dl_pagesize)));
      errstring = _mmap_remain_zero_page(zeropage, zeroend, c->prot, hp_flag);
    }
  return errstring;
}

/*
 * mmap as fixed addr, if the middle part is 2MB aligned,
 * this part should be mmaped in 2MB aligned, else in 4KB aligned
 * 2MB hugepage area should be set with correct permissions, no need to remap
 */
static __always_inline const char *
_mmap_segment_filesz(struct link_map *l, const struct loadcmd *c, ElfW(Addr) mapstart,
              size_t extra_len, int fd, char hp_flag)
{
  void *map_addr = 0;

  size_t relro_len = _get_relro_len(l, c);
  if (c->mapend + extra_len < c->mapstart + relro_len)
      return DL_MAP_SEGMENTS_GNU_RELRO;

  if (relro_len > 0)
    {
      if (__glibc_unlikely (GLRO(dl_debug_mask) & DL_DEBUG_FILES))
        _dl_debug_printf("\t\tmmap relro: [%lx-%lx)\n", mapstart, mapstart + relro_len);
      /*
       * relro part must be mapped as normal page size to avoid
       * _dl_protect_relro failure
       */
      map_addr =  __mmap((void *)mapstart, relro_len, c->prot,
                          MAP_PRIVATE|MAP_FIXED|MAP_FILE,
                          fd, c->mapoff);
      if (__glibc_unlikely (map_addr == MAP_FAILED))
        return DL_MAP_SEGMENTS_ERROR_MAP_SEGMENT;

      mapstart += relro_len;
    }

  size_t prev_map_len = ALIGN_UP(mapstart, SIZE_2MB) - mapstart;
  size_t len = (c->mapend + extra_len) - (c->mapstart + relro_len);
  if (len == 0)
    {
      if (__glibc_unlikely (GLRO(dl_debug_mask) & DL_DEBUG_FILES))
          _dl_debug_printf("\t\tAfter mmapping relro len, remain filesz is zero\n");
      return NULL;
    }
  if (len <= prev_map_len || len - prev_map_len < SIZE_2MB)
    {
      if (__glibc_unlikely (GLRO(dl_debug_mask) & DL_DEBUG_FILES))
        _dl_debug_printf("\t\tmmap all: [%lx-%lx), which includes prev_map_len(0x%lx)\n",
                        mapstart, mapstart + len, prev_map_len);
      mapstart = (ElfW(Addr))__mmap((void *)mapstart, len, c->prot,
                                    MAP_PRIVATE|MAP_FIXED|MAP_FILE,
                                    fd, c->mapoff + relro_len);
      if (__glibc_unlikely ((void *)mapstart == MAP_FAILED))
        return DL_MAP_SEGMENTS_ERROR_MAP_SEGMENT;
      return NULL;
    }

  if (prev_map_len > 0)
    {
      if (__glibc_unlikely (GLRO(dl_debug_mask) & DL_DEBUG_FILES))
        _dl_debug_printf("\t\tmmap prev_map_len: [%lx-%lx)\n",
                         mapstart, mapstart + prev_map_len);
      mapstart = (ElfW(Addr))__mmap((void *)mapstart, prev_map_len, c->prot,
                        MAP_PRIVATE|MAP_FIXED|MAP_FILE,
                        fd, c->mapoff + relro_len);
      if (__glibc_unlikely ((void *)mapstart == MAP_FAILED))
        return DL_MAP_SEGMENTS_ERROR_MAP_SEGMENT;

      map_addr = map_addr == 0 ? (void *)mapstart : map_addr;
      mapstart += prev_map_len;
      len -= prev_map_len;
    }

  size_t mod = len % SIZE_2MB;
  if (__glibc_unlikely (GLRO(dl_debug_mask) & DL_DEBUG_FILES))
    _dl_debug_printf("\t\tmmap hugepage: [%lx-%lx)\n", mapstart, mapstart + len - mod);

  if (hp_flag == DL_HUGEPAGE_LARGE_IN)
  {
    mapstart = (ElfW(Addr))__mmap((void *)mapstart, len - mod, c->prot,
                      		  MAP_FIXED|MAP_PRIVATE|MAP_ANONYMOUS|MAP_HUGETLB|(SHFIT_2MB << MAP_HUGE_SHIFT),
	                          -1, 0);

    if (__glibc_unlikely ((void *)mapstart == MAP_FAILED))
      return DL_MAP_SEGMENTS_ERROR_MAP_SEGMENT;
  } else {
    mapstart = (ElfW(Addr))__mmap((void *)mapstart, len - mod, c->prot,
                                  MAP_FIXED|MAP_PRIVATE|MAP_ANONYMOUS, -1, 0);

    if (__glibc_unlikely ((void *)mapstart == MAP_FAILED))
      return DL_MAP_SEGMENTS_ERROR_MAP_SEGMENT;

    (void) madvise((void *)mapstart, len - mod, MADV_HUGEPAGE);
  }

  if ((c->prot & PROT_WRITE) == 0 && __mprotect((void *)mapstart, len - mod, c->prot | PROT_WRITE) < 0)
    {
      return DL_MAP_SEGMENTS_ERROR_MPROTECT;
    }

  /* Read the segment contents from the file. */
  size_t file_len = (size_t)(c->dataend - c->mapstart) <= prev_map_len + relro_len ? 0 :
                    (size_t)(c->dataend - c->mapstart) - prev_map_len - relro_len;
  if (file_len > 0)
    {
      lseek(fd, c->mapoff + relro_len + prev_map_len, SEEK_SET);
      if ( __read(fd, (void *)mapstart, file_len < len - mod ? file_len : len - mod) < 0)
        return DL_MAP_SEGMENTS_ERROR_READ_SEGMENT;
    }

  if ((c->prot & PROT_WRITE) == 0 &&  __mprotect((void *)mapstart, len - mod, c->prot) < 0)
    {
      return DL_MAP_SEGMENTS_ERROR_MPROTECT;
    }

  map_addr = map_addr == 0 ? (void *)mapstart : map_addr;
  mapstart += len - mod;

  if (__glibc_unlikely (extra_len > 0 && mod > 0))
    return DL_MAP_SEGMENT_ERROR_EXTRA_SIZE;

  if (mod > 0 && __glibc_unlikely (GLRO(dl_debug_mask) & DL_DEBUG_FILES))
    _dl_debug_printf("\t\tmmap tail part: [%lx-%lx)\n", mapstart, mapstart + mod);
  if (mod > 0 && __mmap((void *)mapstart, mod, c->prot,
                MAP_PRIVATE|MAP_FIXED|MAP_FILE,
                fd, c->mapoff + relro_len + prev_map_len + len - mod)
              == MAP_FAILED)
    {
      return DL_MAP_SEGMENTS_ERROR_MAP_SEGMENT;
    }
  return NULL;
}

/*
 * mmap segment filesz tail part only covers the very first part of hugepage,
 * if the size of this tail part reach the threshold, map the tail part in hugepage
 *
 * The tail part must be calculated by mapend, because this is file mmaping,
 * if tail part is calculated by allocend, it will mmap invalid data in file
 * s: mapstart mp: mapend ac: allocend
 * 1. [s, mp) can not cover the tail hugepage start, mp, s, ac are all in same hugepage, no extra space
 *    s mp ac
 *    | |  |
 * |--------|--------|
 *
 * 2. [s, mp) can not cover the tail hugepage start, ac is in the behind hugepage, no extra space
 *    s mp            ac
 *    | |             |
 * |--------|--------|--------|
 *
 * 3. [s, mp) covers the tail hugepage start, mp and the ac in the same hugepage,
 * if (ac - ALIGN_DOWN(mp, SIZE_2MB) < threshold, no extra space; else extra space
 * [mp, ALIGN_UP(mp, SIZE_2MB) which contains initialized data area and hole
 * if ac == mp, the extra space only contains hole
 *      s1           s2  mp ac
 *      |            |   |  |
 * |--------|--------|--------|
 *
 * 4. [s, mp) covers the tail hugepage start, ac is in the behind hugepage,
 * the extra space is [mp, ALIGN_UP(mp, SIZE_2MB) which only contains initialized data area
 *    s1             s2  mp               ac
 *    |              |   |                |
 * |--------|--------|--------|--------|--------|
 *
 * 5. if mp is 2MB aligned, no matter [s, mp) covers the tail hugepage start or not,
 * no extra area
 *    s1             s2  s3   mp           ac
 *    |              |   |    |            |
 * |--------|--------|--------|--------|--------|
 *
 * there are a few points to note:
 * 1. the extra part shold not overlap with the next segment
 * 2. PT_LOAD segment which contains relro section should update mapstart
 */
static __always_inline size_t
_extra_mmap(struct link_map *l, const struct loadcmd loadcmds[], size_t nloadcmds,
            const struct loadcmd *c, ElfW(Addr) mapstart)
{
  ElfW(Addr) mapend = mapstart + (c->mapend - c->mapstart);
  ElfW(Addr) hugepage = ALIGN_DOWN(mapend, SIZE_2MB);
  size_t relro_len = _get_relro_len(l, c);
  mapstart += relro_len;

  /*
   * 1. mapend is 2MB aligned
   * 2. [mapstart, mapend) does not cover the tail hugepage start
   */
  if (mapend == ALIGN_UP(mapend, SIZE_2MB) || mapstart > hugepage)
    return 0;

  /* the initialized data area end in the tail hugepage */
  ElfW(Addr) end = (mapstart - relro_len) + ALIGN_UP(c->allocend - c->mapstart, GLRO(dl_pagesize)) >=
                    ALIGN_UP(mapend, SIZE_2MB) ? ALIGN_UP(mapend, SIZE_2MB) :
                    (mapstart - relro_len) + ALIGN_UP(c->allocend - c->mapstart, GLRO(dl_pagesize));

  size_t extra_len = ALIGN_UP(mapend, SIZE_2MB) - mapend;
  if ((end - hugepage < THRESHOLD * GLRO(dl_pagesize)) || ((c < loadcmds + (nloadcmds - 1)) &&
      (ALIGN_UP(mapend, SIZE_2MB) > (mapstart - relro_len) + c[1].mapstart - c->mapstart)))
    {
      extra_len = 0;
    }

  return extra_len;
}

static __always_inline const char *
__mmap_segment_normalsz(const struct loadcmd *c, ElfW(Addr) mapstart, int fd,
        size_t *mapseglen)
{
  if (__glibc_unlikely (GLRO(dl_debug_mask) & DL_DEBUG_FILES))
    _dl_debug_printf("\t\t%s mmap current segment in normal page:[%lx-%lx)\n", __func__, mapstart,
            mapstart + (ALIGN_UP(c->allocend, GLRO(dl_pagesize)) - c->mapstart));

  if (c->mapend > c->mapstart &&
          (__mmap((void *)mapstart, c->mapend - c->mapstart, c->prot,
                  MAP_FILE|MAP_PRIVATE|MAP_FIXED, fd, c->mapoff) == MAP_FAILED))
    return DL_MAP_SEGMENTS_ERROR_MAP_SEGMENT;

  if (c->allocend > c->dataend) {
    ElfW(Addr) zero, zeroend, zeropage;

    zero = mapstart + c->dataend - c->mapstart;
    zeroend = mapstart + c->allocend - c->mapstart;
    zeropage = ((zero + GLRO(dl_pagesize) - 1)
                & ~(GLRO(dl_pagesize) - 1));

    if (zeroend < zeropage)
      zeropage = zeroend;

    if (zeropage > zero) {
      if (__glibc_unlikely ((c->prot & PROT_WRITE) == 0)) {
        if (__mprotect ((caddr_t) (zero & ~(GLRO(dl_pagesize) - 1)),
                GLRO(dl_pagesize), c->prot|PROT_WRITE) < 0)
          return DL_MAP_SEGMENTS_ERROR_MPROTECT;
      }

      memset ((void *) zero, '\0', zeropage - zero);

      if (__glibc_unlikely ((c->prot & PROT_WRITE) == 0))
        __mprotect ((caddr_t) (zero & ~(GLRO(dl_pagesize) - 1)),
                GLRO(dl_pagesize), c->prot);
    }

    if (zeroend > zeropage) {
      if (__mmap ((caddr_t) zeropage, zeroend - zeropage, c->prot,
                  MAP_ANON|MAP_PRIVATE|MAP_FIXED, -1, 0) == MAP_FAILED)
        return DL_MAP_SEGMENTS_ERROR_MAP_ZERO_FILL;
    }
  }

  *mapseglen = ALIGN_UP(c->allocend, GLRO(dl_pagesize)) - c->mapstart;
  return NULL;
}

/*
 * PT_LOAD segment is described by p_filesz and p_memsz.
 * The bytes from the file are mapped to the beginning of the memory segment.
 * If the segment’s memory size (p_memsz) is larger than the file size (p_filesz),
 * the extra bytes are defined to hold the value 0 and to follow the segment’s
 * initialized area
 */
static __always_inline const char *
_mmap_segment(struct link_map *l, const struct loadcmd loadcmds[], size_t nloadcmds,
              const struct loadcmd *c, ElfW(Addr) mapstart, int fd,
              size_t *mapseglen, const char hp_bitmap[])
{
  const char * errstring = NULL;
  size_t extra_len, memsz_len = 0;
  char hp_flag;

  if (__glibc_unlikely (GLRO(dl_debug_mask) & DL_DEBUG_FILES))
    _dl_debug_printf("\t%s mmap from %lx: \n\t{\n", __func__, mapstart);

  hp_flag = hp_bitmap[((void *)c - (void *)loadcmds) / sizeof(struct loadcmd)];
  if (!hp_flag)
    {
      errstring = __mmap_segment_normalsz(c, mapstart, fd, mapseglen);
      if (__glibc_unlikely (GLRO(dl_debug_mask) & DL_DEBUG_FILES))
        _dl_debug_printf("\t} => mapseglen = 0x%lx\n", *mapseglen);
      return errstring;
    }

  extra_len = _extra_mmap(l, loadcmds, nloadcmds, c, mapstart);
  errstring = _mmap_segment_filesz(l, c, mapstart, extra_len, fd, hp_flag);
  if (__glibc_unlikely (errstring != NULL))
    return errstring;
  errstring = _mmap_segment_memsz(l, c, mapstart, extra_len, &memsz_len, hp_flag);
  if (__glibc_unlikely (errstring != NULL))
    return errstring;

  *mapseglen = c->mapend - c->mapstart + extra_len + memsz_len;
  if (__glibc_unlikely (GLRO(dl_debug_mask) & DL_DEBUG_FILES))
    _dl_debug_printf("\t} => mapseglen = 0x%lx, extra_len = 0x%lx, memsz_len = 0x%lx\n",
                     *mapseglen, extra_len, memsz_len);
  return NULL;
}

static __always_inline const char *
_dl_map_segments_largein (struct link_map *l, int fd,
                  const ElfW(Ehdr) *header, int type,
                  const struct loadcmd loadcmds[], size_t nloadcmds,
                  const size_t maplength, const char hp_bitmap[])
{
  if (__glibc_unlikely (type != ET_DYN))
    return DL_MAP_SEGMENTS_ERROR_TYPE;

  const char *errstring = NULL;
  const struct loadcmd *text = _find_exec_segment(loadcmds, nloadcmds);
  if (__glibc_unlikely (text == NULL))
    return DL_FIND_EXEC_SEGMENT_ERROR;

  size_t maparealen;
  void *map_area_start = __mmap_reserved_area(loadcmds, nloadcmds, &maparealen);
  if (__glibc_unlikely (map_area_start == MAP_FAILED))
    return DL_MAP_RESERVED_HUGEPAGE_AREA_ERROR;
  if (__glibc_unlikely (GLRO(dl_debug_mask) & DL_DEBUG_FILES))
    _dl_debug_printf("reserved area:[%lx-%lx)\n",
      (unsigned long)map_area_start, (unsigned long)map_area_start + maparealen);

  /* First to mmap text segment */
  const struct loadcmd * c = loadcmds;
  ElfW(Addr) text_addr = ALIGN_UP((ElfW(Addr))map_area_start + (text->mapstart - c->mapstart), SIZE_2MB);
  size_t mapseglen = 0;
  errstring = _mmap_segment(l, loadcmds, nloadcmds, text, text_addr, fd,
          &mapseglen, hp_bitmap);
  if (__glibc_unlikely(errstring != NULL))
    goto unmap_reserved_area;

  const struct loadcmd *prev = text;
  c = text + 1;
  ElfW(Addr) map_addr = text_addr;
  while (c < &loadcmds[nloadcmds])
    {
      if (prev->mapstart + mapseglen > c->mapstart || c->mapstart < prev->mapstart)
        {
          errstring = DL_MAP_SEGMENTS_ERROR_ARRANGE;
          goto unmap_reserved_area;
        }

      map_addr += c->mapstart - prev->mapstart;
      errstring = _mmap_segment(l, loadcmds, nloadcmds, c, map_addr, fd,
              &mapseglen, hp_bitmap);
      if (__glibc_unlikely(errstring != NULL))
        goto unmap_reserved_area;
      prev = c;
      ++c;
    }
  ElfW(Addr) l_map_end = map_addr + mapseglen;

  /* search for the first segment */
  prev = text;
  c = text - 1;
  map_addr = text_addr;
  while (c >= loadcmds)
    {
      if (prev->mapstart < c->mapstart)
        {
          errstring = DL_MAP_SEGMENTS_ERROR_ARRANGE;
          goto unmap_reserved_area;
        }

      map_addr -= prev->mapstart - c->mapstart;
      errstring = _mmap_segment(l, loadcmds, nloadcmds, c, map_addr, fd,
              &mapseglen, hp_bitmap);
      if (__glibc_unlikely(errstring != NULL))
        goto unmap_reserved_area;

      if (c->mapstart + mapseglen > prev->mapstart)
        {
          errstring = DL_MAP_SEGMENTS_ERROR_ARRANGE;
          goto unmap_reserved_area;
        }

      prev = c;
      --c;
    }

  ++c;
  l->l_map_start = map_addr;
  l->l_map_end = l->l_map_start + maplength;
  l->l_addr = l->l_map_start - c->mapstart;
  l->l_contiguous = 1;

  c = loadcmds;
  while (c < &loadcmds[nloadcmds])
    {
      _dl_postprocess_loadcmd (l, header, c);
      ++c;
    }

  if (l->l_map_start > (ElfW(Addr))map_area_start)
    {
      if (__glibc_unlikely (GLRO(dl_debug_mask) & DL_DEBUG_FILES))
        _dl_debug_printf("__munmap head reserved area [%lx-%lx)\n",  (ElfW(Addr))map_area_start, l->l_map_start);
      __munmap(map_area_start, l->l_map_start - (ElfW(Addr))map_area_start);
    }

  /*
   * l->l_map_end is caculated by maplength, l_map_end may end with extra space
   * use l->l_map_end may munmap extra space part
   */
  if ((ElfW(Addr))map_area_start + maparealen > l_map_end)
    {
      if (__glibc_unlikely (GLRO(dl_debug_mask) & DL_DEBUG_FILES))
        _dl_debug_printf("__munmap tail reserved area [%lx-%lx)\n", l_map_end, (ElfW(Addr))map_area_start + maparealen);
      __munmap((void *)l_map_end, (ElfW(Addr))map_area_start + maparealen - l_map_end);
    }

  return NULL;

unmap_reserved_area:
  __munmap(map_area_start, maparealen);

  return errstring;
}
