/* Mark ELF object ELF header hugepage flag Generic version.
   Copyright (C) 2021-2021 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <http://www.gnu.org/licenses/>.  */
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <elf.h>
#include <link.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/types.h>
#include "tst-get-ephdr.h"

#define TOOL_NAME "hugepageedit"

void print_usage(void)
{
  fprintf(stderr, "%s [-x] [-d] [-i index] <ELF file>\n"  \
        "\tdefault mark all PT_LOAD segment PF_HUGEPAGE flag\n"  \
        "\t-x option only mark executable PT_LOAD segment PF_HUGEPAGE flag\n" \
        "\t-i [index(start from 0)] option specifies the index that marks the PT_LOAD segment PF_HUGEPAGE flag\n" \
        "\t-d option delete all the PT_LOAD segment PF_HUGEPAGE flag\n", TOOL_NAME);
}


static long parse_index(char *str)
{
    char *endptr;

    errno = 0;
    long val = strtol(str, &endptr, 10);

    if (errno != 0) {
        perror("strtol");
        return -1;
    }

    if (endptr == str) {
        fprintf(stderr, "No digits were found in -i option\n");
        return -1;
    }

    if (*endptr != '\0') {
        fprintf(stderr, "Invalid characters %s in -i %s option\n", endptr, str);
        return -1;
    }

    if (val < 0) {
        fprintf(stderr, "Negative index %ld in -i %s option\n", val, str);
        return -1;
    }

    return val;
}


int main(int argc, char *argv[])
{
  size_t length;
  int exit_status = -1;
  int i, opt, delete = 0, exec_only = 0, index_set = 0;
  long index = -1, index_count = 0;
  while ((opt = getopt(argc, argv, "dxi:")) != -1)
    {
      switch (opt)
        {
          case 'd':
            delete = 1;
            break;
          case 'x':
            exec_only = 1;
            break;
          case 'i':
            index = parse_index(optarg);
            index_set = 1;
            if (index < 0)
                return -1;
            index_count++;
            if (index_count > 1) {
              fprintf(stderr, "too many -i options in command line\n");
              return -1;
            }
            break;
          default:
            print_usage();
            return 0;
        }
    }

  if (delete + exec_only + index_set > 1)
    {
      fprintf(stderr, "can not specify -x, -d and -i option at the same time\n");
      return -1;
    }

  if (optind >= argc)
    {
      fprintf(stderr, "Expected argument after options\n");
      return -1;
    }

  int fd = open(argv[optind], O_RDWR);
  if (fd < 0)
    {
      perror("open");
      return -1;
    }

  void *ehdr = get_ehdr(fd, &length);
  if (ehdr == NULL)
    goto close_fd;

  if (index_set && index >= ((ElfW(Ehdr) *)ehdr)->e_phnum) {
      fprintf(stderr, "Index %ld in -i %s option out of PT_LOAD segment range\n",
              index, argv[optind]);
      goto close_fd;
  }

  ElfW(Phdr) *phdr = (ElfW(Phdr) *)get_phdr(ehdr, length);
  if (phdr == NULL)
    goto unmap;
  /*
   * Here, mark hugepage flag in ELF header e_ident padding bytes won't work.
   * elf/dl-load.c open_verify will check if shared object ELF header e_ident
   * padding bytes match expected[EI_NIDENT] byte array which padding bytes
   * should be zero. If it mismatches, ld.so will exit abnormally
   */
  for (i = 0; i < ((ElfW(Ehdr) *)ehdr)->e_phnum; i++)
    {
      if (phdr[i].p_type == PT_LOAD)
        {
          if (delete)
            {
              phdr[i].p_flags &= ~PF_HUGEPAGE;
            }
          else
            {
              if ((exec_only && !(phdr[i].p_flags & PF_X)) || (index_set && index != i))
                continue;
              phdr[i].p_flags |= PF_HUGEPAGE;
            }
        } else {
          if (index_set && index == i) {
            fprintf(stderr, "Index %ld in -i %s option is not a PT_LOAD segment\n",
                    index, argv[optind]);
            goto unmap;
          }
        }
    }
  exit_status = 0;

unmap:
  munmap(ehdr, length);

close_fd:
  close(fd);

  return exit_status;
}
