/* mmap ELF file return ELF header and get mmap length
   Copyright (C) 1995-2021 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <http://www.gnu.org/licenses/>.  */
#include <stdio.h>
#include <elf.h>
#include <link.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/mman.h>

/* reference kernel load_elf_phdrs program header table size constraint */
#define ELF_MIN_ALIGN 4096

static __always_inline void *get_ehdr(int fd, size_t *len)
{
  struct stat statbuf;
  *len = 0;
  if (fstat(fd, &statbuf) != 0)
    {
      perror("fstat");
      return NULL;
    }

  /* this ensures file is large enough to hold ELF header */
  if (statbuf.st_size < sizeof (ElfW(Ehdr)))
    {
      fprintf(stderr, "file is not large enough to hold ELF header\n");
      return NULL;
    }

  void *ehdr = mmap(NULL, statbuf.st_size, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);
  if (ehdr == MAP_FAILED)
    {
      perror("mmap");
      return NULL;
    }

  if (memcmp(((ElfW(Ehdr) *) ehdr)->e_ident, ELFMAG, SELFMAG) != 0)
    {
      fprintf(stderr, "file is not ELF format\n");
      munmap(ehdr, statbuf.st_size);
      ehdr = NULL;
    }

  *len = statbuf.st_size;
  return ehdr;
}

static __always_inline int check_ptr(void *ptr, void *start, size_t len)
{
  if (ptr < start || ptr > start + len)
    return -1;
  return 0;
}

static __always_inline void *get_phdr(void *ehdr, size_t length)
{
  if (((ElfW(Ehdr) *)ehdr)->e_phentsize != sizeof(ElfW(Phdr)))
    {
      fprintf(stderr, "ELF header's e_phentsize mismatch ElfW(Phdr) size\n");
      return NULL;
    }

  unsigned int size = ((ElfW(Ehdr) *)ehdr)->e_phnum * sizeof(ElfW(Phdr));
  if (size == 0 || size > ELF_MIN_ALIGN)
    {
      fprintf(stderr, "The program header table size specified by ELF header is abnormal: %u\n", size);
      return NULL;
    }

  void *ephdr_s = ehdr + ((ElfW(Ehdr) *)ehdr)->e_phoff;
  void *ephdr_e = ehdr + ((ElfW(Ehdr) *)ehdr)->e_phoff + size;

  if (check_ptr(ephdr_s, ehdr, length) ||
        check_ptr(ephdr_e, ehdr, length))
    {
      fprintf(stderr, "ELF porgram header table is not fully mmaped\n");
      return NULL;
    }

  return ephdr_s;
}
